package king.logical;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class LogicalObject {
	
	public enum Type{Literal,Conjunction,Disjunction,Implication,Biconditional};
	protected Type type;
	protected List<LogicalObject> subObjects;
	protected Literal literal;

	
	protected LogicalObject(){
		
	}
	
	public boolean isLiteral(){
		return type == Type.Literal;
	}
	
	public boolean isConjunction(){
		return type == Type.Conjunction;
	}
	
	public boolean isDisjunction(){
		return type == Type.Disjunction;
	}
	
	public boolean isImplication(){
		return type == Type.Implication;
	}
	
	public boolean isBiconditional(){
		return type == Type.Biconditional;
	}
	
	public final Type getType(){
		return type;
	}
	
	public static LogicalObject newLiteral(String name,Boolean value){
		LogicalObject lo = new LogicalObject();
		lo.type = Type.Literal;
		lo.literal = new Literal(name,value);
		return lo;
	}
	
	
	public static LogicalObject newConjunction(LogicalObject[] list){
		LogicalObject lo = new LogicalObject();
		lo.type = Type.Conjunction;
		lo.subObjects = Arrays.asList(list);
		
		return lo;
	}
	
	
	public static LogicalObject newDisjunction(LogicalObject[] list){
		LogicalObject lo = new LogicalObject();
		lo.type = Type.Disjunction;
		lo.subObjects = Arrays.asList(list);
		return lo;
	}
	
	public static LogicalObject newImplication(LogicalObject left, LogicalObject right){
		LogicalObject lo = new LogicalObject();
		lo.type = Type.Implication;
		lo.subObjects = new ArrayList<LogicalObject>();
		lo.subObjects.add(left);
		lo.subObjects.add(right);
		return lo;
	}
	
	public static LogicalObject newBiconditional(LogicalObject left, LogicalObject right){
		LogicalObject lo = new LogicalObject();
		lo.type = Type.Biconditional;
		lo.subObjects = new ArrayList<LogicalObject>();
		lo.subObjects.add(left);
		lo.subObjects.add(right);
		return lo;
	}
	
	public LogicalObject getLeft(){
		if(type != Type.Implication && type != Type.Biconditional)
			return null;
		return subObjects.get(0);
	}
	
	public LogicalObject getRight(){
		if(type != Type.Implication && type != Type.Biconditional)
			return null;
		return subObjects.get(1);
	}
	
	public String getName(){
		if(type != Type.Literal)
			return "ERROR: NON_LITERAL";
		return literal.symbol;
	}
	
	public Boolean getValue(){
		if(type != Type.Literal)
			return false;
		return literal.value;
	}
	
	public List<ClauseCNF> convertToCNF(){
		if(isLiteral()){
			return Arrays.asList(new ClauseCNF(literal.symbol,literal.value));
		}
		if(isConjunction()){
			List<ClauseCNF> clauses = new ArrayList<ClauseCNF>();
			for(LogicalObject lo : subObjects){
				clauses.addAll(lo.convertToCNF());
			}
			return clauses;
		}
		if(isDisjunction()){
			List<ClauseCNF> clauses = null;
			for(LogicalObject lo : subObjects)
				clauses = (clauses==null)?lo.convertToCNF():merge2CNF(clauses,lo.convertToCNF());
			return clauses;
		}
		if(isImplication()){
			List<ClauseCNF> right = getRight().convertToCNF();
			List<ClauseCNF> left = getInverted(getLeft().convertToCNF());
			return merge2CNF(left,right);
		}
		List<ClauseCNF> right = getRight().convertToCNF();
		List<ClauseCNF> left = getLeft().convertToCNF();
		List<ClauseCNF> clauses = merge2CNF(left,getInverted(right));
		clauses.addAll(merge2CNF(getInverted(left),right));
		return clauses;
	}
	
	private static List<ClauseCNF> getInverted(List<ClauseCNF> list){
		List<ClauseCNF> clauses = null;
		List<List<ClauseCNF>> newCNFs = new ArrayList<List<ClauseCNF>>();
		for(ClauseCNF clause : list)
			newCNFs.add(clause.deMorgans());
		for(List<ClauseCNF> cnf : newCNFs)
			clauses = (clauses==null)?cnf:merge2CNF(clauses,cnf);
		return clauses;
	}
	
	private static List<ClauseCNF> merge2CNF(List<ClauseCNF> a,List<ClauseCNF> b){
		List<ClauseCNF> merged = new ArrayList<ClauseCNF>();
		for(ClauseCNF cA : a)
			for(ClauseCNF cB : b)
				if(cA.absorb(cB))
					merged.add(cA);
		return merged.isEmpty()?null:merged;
	}

}
